import ast
import dataclasses
import inspect
from dataclasses import dataclass
from enum import Enum
from functools import cache
from pathlib import Path
from types import ModuleType
from typing import Any, Generic, Protocol, TypeAlias, TypeVar, cast

from pydantic import BaseModel
from typing_extensions import Self

from cadwyn._compat import PydanticFieldWrapper, model_fields
from cadwyn._package_utils import IdentifierPythonPath
from cadwyn.exceptions import CodeGenerationError
from cadwyn.structure.versions import Version

_FieldName: TypeAlias = str
_CodegenPluginASTType = TypeVar("_CodegenPluginASTType", bound=ast.AST)


@dataclasses.dataclass(slots=True)
class PydanticModelWrapper:
    cls: type[BaseModel]
    name: str
    fields: dict[_FieldName, PydanticFieldWrapper]
    _parents: list[Self] | None = dataclasses.field(init=False, default=None)

    def _get_parents(self, schemas: "dict[IdentifierPythonPath, Self]"):
        if self._parents is not None:
            return self._parents
        parents = []
        for base in self.cls.mro()[1:]:
            schema_path = f"{base.__module__}.{base.__name__}"

            if schema_path in schemas:
                parents.append(schemas[schema_path])
            elif issubclass(base, BaseModel):
                parents.append(type(self)(base, base.__name__, get_fields_from_model(base)))
        self._parents = parents
        return parents

    def _get_defined_fields(self, schemas: "dict[IdentifierPythonPath, Self]") -> dict[str, PydanticFieldWrapper]:
        fields = {}

        for parent in reversed(self._get_parents(schemas)):
            fields |= parent.fields

        return fields | self.fields


@cache
def get_fields_from_model(cls: type) -> dict[str, PydanticFieldWrapper]:
    if not isinstance(cls, type) or not issubclass(cls, BaseModel):
        raise CodeGenerationError(f"Model {cls} is not a subclass of BaseModel")

    fields = model_fields(cls)
    try:
        source = inspect.getsource(cls)
    except OSError:
        return {
            field_name: PydanticFieldWrapper(
                annotation=field.annotation,
                init_model_field=field,
            )
            for field_name, field in fields.items()
        }
    else:
        cls_ast = cast(ast.ClassDef, ast.parse(source).body[0])
        return {
            node.target.id: PydanticFieldWrapper(
                annotation=fields[node.target.id].annotation,
                init_model_field=fields[node.target.id],
                annotation_ast=node.annotation,
                field_ast=node.value,
            )
            for node in cls_ast.body
            if isinstance(node, ast.AnnAssign) and isinstance(node.target, ast.Name) and node.target.id in fields
        }


@dataclass(slots=True)
class _EnumWrapper:
    cls: type[Enum]
    members: dict[_FieldName, Any]


@dataclass(slots=True)
class _ModuleWrapper:
    value: ModuleType
    extra_imports: list[ast.Import | ast.ImportFrom] = dataclasses.field(default_factory=list)


@dataclasses.dataclass(slots=True, kw_only=True)
class GlobalCodegenContext:
    current_version: Version
    latest_version: Version = dataclasses.field(init=False)
    versions: list[Version]
    schemas: dict[IdentifierPythonPath, PydanticModelWrapper] = dataclasses.field(repr=False)
    enums: dict[IdentifierPythonPath, _EnumWrapper] = dataclasses.field(repr=False)
    modules: dict[IdentifierPythonPath, _ModuleWrapper] = dataclasses.field(repr=False)
    extra: dict[str, Any]

    def __post_init__(self):
        self.latest_version = max(self.versions, key=lambda v: v.value)

    @property
    def current_version_is_latest(self):
        return self.latest_version == self.current_version


@dataclasses.dataclass(slots=True, kw_only=True)
class CodegenContext(GlobalCodegenContext):
    # This attribute is extremely useful for calculating relative imports
    index_of_latest_package_dir_in_module_python_path: int
    module_python_path: str
    module_path: Path
    template_module: ModuleType
    all_names_defined_on_toplevel_of_file: dict[IdentifierPythonPath, str]


class CodegenPlugin(Protocol, Generic[_CodegenPluginASTType]):
    @property
    def node_type(self) -> type[_CodegenPluginASTType]:
        raise NotImplementedError

    @staticmethod
    def __call__(node: _CodegenPluginASTType, context: CodegenContext) -> _CodegenPluginASTType:
        raise NotImplementedError


class MigrationPlugin(Protocol):
    @staticmethod
    def __call__(context: GlobalCodegenContext) -> None:
        raise NotImplementedError



import importlib
import sys
from pathlib import Path
from typing import Any

import typer

from cadwyn.structure.versions import VersionBundle

app = typer.Typer(
    name="cadwyn",
    add_completion=False,
    help="Modern Stripe-like API versioning in FastAPI",
)


def version_callback(value: bool):
    if value:
        from . import __version__

        typer.echo(f"Cadwyn {__version__}")
        raise typer.Exit


@app.command(name="generate-code-for-versioned-packages")
def generate_versioned_packages(
    path_to_template_package: str = typer.Argument(
        ...,
        help=(
            "The python path to the template package, from which we will generate the versioned packages. "
            "Format: 'path.to.template_package'"
        ),
        show_default=False,
    ),
    full_path_to_version_bundle: str = typer.Argument(
        ...,
        help="The python path to the version bundle. Format: 'path.to.version_bundle:my_version_bundle_variable'",
        show_default=False,
    ),
    ignore_coverage_for_latest_aliases: bool = typer.Option(
        default=True,
        help="Add a pragma: no cover comment to the star imports in the generated version of the latest module.",
    ),
) -> None:
    """For each version in the version bundle, generate a versioned package based on the template package"""
    from .codegen._main import generate_code_for_versioned_packages

    sys.path.append(str(Path.cwd()))
    template_package = importlib.import_module(path_to_template_package)
    path_to_version_bundle, version_bundle_variable_name = full_path_to_version_bundle.split(":")
    version_bundle_module = importlib.import_module(path_to_version_bundle)
    possibly_version_bundle = getattr(version_bundle_module, version_bundle_variable_name)
    version_bundle = _get_version_bundle(possibly_version_bundle)

    return generate_code_for_versioned_packages(
        template_package,
        version_bundle,
        ignore_coverage_for_latest_aliases=ignore_coverage_for_latest_aliases,
    )


def _get_version_bundle(possibly_version_bundle: Any) -> VersionBundle:
    if not isinstance(possibly_version_bundle, VersionBundle):
        err = TypeError(
            "The provided version bundle is not a version bundle and "
            "is not a zero-argument callable that returns the version bundle. "
            f"Instead received: {possibly_version_bundle}",
        )
        if callable(possibly_version_bundle):
            try:
                return _get_version_bundle(possibly_version_bundle())
            except TypeError as e:
                raise err from e
        raise err
    return possibly_version_bundle


@app.callback()
def main(
    version: bool = typer.Option(None, "-V", "--version", callback=version_callback, is_eager=True),
):
    ...


if __name__ == "__main__":
    app()



import ast
import inspect
import re
from collections.abc import Callable
from enum import Enum, auto
from pathlib import Path
from types import GenericAlias, LambdaType, ModuleType, NoneType
from typing import (  # noqa: UP035
    TYPE_CHECKING,
    Any,
    List,
    get_args,
    get_origin,
)

from cadwyn._compat import (
    PYDANTIC_V2,
    get_attrs_that_are_not_from_field_and_that_are_from_field,
    is_pydantic_constrained_type,
)
from cadwyn._package_utils import (
    get_absolute_python_path_of_import,
)
from cadwyn._utils import PlainRepr, UnionType
from cadwyn.exceptions import CodeGenerationError, InvalidGenerationInstructionError, ModuleIsNotAvailableAsTextError

if TYPE_CHECKING:
    import annotated_types

_LambdaFunctionName = (lambda: None).__name__  # pragma: no branch
_RE_CAMEL_TO_SNAKE = re.compile(r"(?<!^)(?=[A-Z])")


# A parent type of typing._GenericAlias
_BaseGenericAlias = type(List[int]).mro()[1]  # noqa: UP006

# type(list[int]) and type(List[int]) are different which is why we have to do this.
# Please note that this problem is much wider than just lists which is why we use typing._BaseGenericAlias
# instead of typing._GenericAlias.
GenericAliasUnion = GenericAlias | _BaseGenericAlias

_LambdaFunctionName = (lambda: None).__name__  # pragma: no branch


def get_fancy_repr(value: Any):
    if PYDANTIC_V2:
        import annotated_types

        if isinstance(value, annotated_types.GroupedMetadata) and hasattr(type(value), "__dataclass_fields__"):
            return transform_grouped_metadata(value)
    if isinstance(value, list | tuple | set | frozenset):
        return transform_collection(value)
    if isinstance(value, dict):
        return transform_dict(value)
    if isinstance(value, GenericAliasUnion):
        return transform_generic_alias(value)
    if value is None or value is NoneType:
        return transform_none(value)
    if isinstance(value, type):
        return transform_type(value)
    if isinstance(value, Enum):
        return transform_enum(value)
    if isinstance(value, auto):
        return transform_auto(value)
    if isinstance(value, UnionType):
        return transform_union(value)
    if isinstance(value, LambdaType) and _LambdaFunctionName == value.__name__:
        return transform_lambda(value)
    if inspect.isfunction(value):
        return transform_function(value)
    else:
        return transform_other(value)


def transform_grouped_metadata(value: "annotated_types.GroupedMetadata"):
    modified_fields = []
    empty_obj = type(value)

    for key in empty_obj.__dataclass_fields__:  # pyright: ignore[reportGeneralTypeIssues]
        if getattr(value, key) != getattr(empty_obj, key):
            modified_fields.append((key, getattr(value, key)))

    return PlainRepr(
        value.__class__.__name__
        + "("
        + ", ".join(f"{PlainRepr(key)}={get_fancy_repr(v)}" for key, v in modified_fields)
        + ")",
    )


def transform_collection(value: list | tuple | set | frozenset) -> Any:
    return PlainRepr(value.__class__(map(get_fancy_repr, value)))


def transform_dict(value: dict) -> Any:
    return PlainRepr(
        value.__class__((get_fancy_repr(k), get_fancy_repr(v)) for k, v in value.items()),
    )


def transform_generic_alias(value: GenericAliasUnion) -> Any:
    return f"{get_fancy_repr(get_origin(value))}[{', '.join(get_fancy_repr(a) for a in get_args(value))}]"


def transform_none(_: NoneType) -> Any:
    return "None"


def transform_type(value: type) -> Any:
    # This is a hack for pydantic's Constrained types
    if is_pydantic_constrained_type(value):
        if get_attrs_that_are_not_from_field_and_that_are_from_field(value)[0]:
            parent = value.mro()[1]
            snake_case = _RE_CAMEL_TO_SNAKE.sub("_", value.__name__)
            cls_name = "con" + "".join(snake_case.split("_")[1:-1])
            return (
                cls_name.lower()
                + "("
                + ", ".join(
                    [
                        f"{key}={get_fancy_repr(val)}"
                        for key, val in value.__dict__.items()
                        if not key.startswith("_") and val is not None and val != parent.__dict__[key]
                    ],
                )
                + ")"
            )
        else:
            # In pydantic V1:
            # MRO of constr looks like: [ConstrainedStrValue, pydantic.types.ConstrainedStr, str, object]
            #                                                                                -2     -1
            #                                                                                ^^^
            value = value.mro()[-2]

    return value.__name__


def transform_enum(value: Enum) -> Any:
    return PlainRepr(f"{value.__class__.__name__}.{value.name}")


def transform_auto(_: auto) -> Any:
    return PlainRepr("auto()")


def transform_union(value: UnionType) -> Any:
    return "typing.Union[" + (", ".join(get_fancy_repr(a) for a in get_args(value))) + "]"


def transform_lambda(value: LambdaType) -> Any:
    # We clean source because getsource() can return only a part of the expression which
    # on its own is not a valid expression such as: "\n  .had(default_factory=lambda: 91)"
    return _get_lambda_source_from_default_factory(inspect.getsource(value).strip(" \n\t."))


def transform_function(value: Callable) -> Any:
    return PlainRepr(value.__name__)


def transform_other(value: Any) -> Any:
    return PlainRepr(repr(value))


def _get_lambda_source_from_default_factory(source: str) -> str:
    found_lambdas: list[ast.Lambda] = []

    ast.parse(source)
    for node in ast.walk(ast.parse(source)):
        if isinstance(node, ast.keyword) and node.arg == "default_factory" and isinstance(node.value, ast.Lambda):
            found_lambdas.append(node.value)
    if len(found_lambdas) == 1:
        return ast.unparse(found_lambdas[0])
    # These two errors are really hard to cover. Not sure if even possible, honestly :)
    elif len(found_lambdas) == 0:  # pragma: no cover
        raise InvalidGenerationInstructionError(
            f"No lambda found in default_factory even though one was passed: {source}",
        )
    else:  # pragma: no cover
        raise InvalidGenerationInstructionError(
            "More than one lambda found in default_factory. This is not supported.",
        )


def read_python_module(module: ModuleType) -> str:
    # Can be cached in the future to gain some speedups
    try:
        return inspect.getsource(module)
    except OSError as e:
        if module.__file__ is None:  # pragma: no cover
            raise CodeGenerationError(f"Failed to get file path to the module {module}") from e
        path = Path(module.__file__)
        if path.is_file() and path.read_text() == "":
            return ""
        raise ModuleIsNotAvailableAsTextError(  # pragma: no cover
            f"Failed to get source code for module {module}. "
            "This is likely because this module is not available as code "
            "(it could be a compiled C extension or even a .pyc file). "
            "Cadwyn does not support models from such code. "
            "Please, open an issue on Cadwyn's issue tracker if you believe that your use case is valid "
            "and if you believe that it is possible for Cadwyn to support it.",
        ) from e


def get_all_names_defined_at_toplevel_of_module(body: ast.Module, module_python_path: str) -> dict[str, str]:
    """Some day we will want to use this to auto-add imports for new symbols in versions. Some day..."""
    defined_names = {}
    for node in body.body:
        if isinstance(node, ast.ClassDef | ast.FunctionDef | ast.AsyncFunctionDef):
            defined_names[node.name] = module_python_path
        elif isinstance(node, ast.Assign):
            for target in node.targets:
                if isinstance(target, ast.Name):
                    defined_names[target.id] = module_python_path
        elif isinstance(node, ast.ImportFrom):
            for name in node.names:
                defined_names[name.name] = get_absolute_python_path_of_import(node, module_python_path)
        elif isinstance(node, ast.Import):
            for name in node.names:
                defined_names[name.name] = name.name
    return defined_names


def add_keyword_to_call(attr_name: str, attr_value: Any, call: ast.Call):
    new_keyword = get_ast_keyword_from_argument_name_and_value(attr_name, attr_value)
    for i, keyword in enumerate(call.keywords):
        if keyword.arg == attr_name:
            call.keywords[i] = new_keyword
            break
    else:
        call.keywords.append(new_keyword)


def get_ast_keyword_from_argument_name_and_value(name: str, value: Any):
    return ast.keyword(
        arg=name,
        value=ast.parse(get_fancy_repr(value), mode="eval").body,
    )


def pop_docstring_from_cls_body(cls_body: list[ast.stmt]) -> list[ast.stmt]:
    if (
        len(cls_body) > 0
        and isinstance(cls_body[0], ast.Expr)
        and isinstance(cls_body[0].value, ast.Constant)
        and isinstance(cls_body[0].value.value, str)
    ):
        return [cls_body.pop(0)]
    else:
        return []



from ._common import CodegenContext, GlobalCodegenContext
from ._main import DEFAULT_CODEGEN_MIGRATION_PLUGINS, DEFAULT_CODEGEN_PLUGINS, generate_code_for_versioned_packages

__all__ = [
    "generate_code_for_versioned_packages",
    "CodegenContext",
    "GlobalCodegenContext",
    "DEFAULT_CODEGEN_PLUGINS",
    "DEFAULT_CODEGEN_MIGRATION_PLUGINS",
]



import ast
from collections.abc import Sequence

from typing_extensions import assert_never

from cadwyn._compat import FieldInfo, PydanticFieldWrapper, dict_of_empty_field_info, is_pydantic_constrained_type
from cadwyn._package_utils import IdentifierPythonPath, get_cls_pythonpath
from cadwyn._utils import Sentinel
from cadwyn.codegen._asts import add_keyword_to_call
from cadwyn.codegen._common import GlobalCodegenContext, PydanticModelWrapper, _EnumWrapper
from cadwyn.exceptions import InvalidGenerationInstructionError
from cadwyn.structure.enums import AlterEnumSubInstruction, EnumDidntHaveMembersInstruction, EnumHadMembersInstruction
from cadwyn.structure.schemas import (
    AlterSchemaInstruction,
    AlterSchemaSubInstruction,
    OldSchemaFieldDidntExist,
    OldSchemaFieldExistedWith,
    OldSchemaFieldHad,
)


def class_migration_plugin(context: GlobalCodegenContext):
    for version_change in context.current_version.version_changes:
        _apply_alter_schema_instructions(
            context.schemas,
            version_change.alter_schema_instructions,
            version_change.__name__,
        )
        _apply_alter_enum_instructions(
            context.enums,
            version_change.alter_enum_instructions,
            version_change.__name__,
        )


def _apply_alter_schema_instructions(
    modified_schemas: dict[IdentifierPythonPath, PydanticModelWrapper],
    alter_schema_instructions: Sequence[AlterSchemaSubInstruction | AlterSchemaInstruction],
    version_change_name: str,
):
    for alter_schema_instruction in alter_schema_instructions:
        schema = alter_schema_instruction.schema
        schema_path = get_cls_pythonpath(schema)
        mutable_schema_info = modified_schemas[schema_path]
        if isinstance(alter_schema_instruction, OldSchemaFieldDidntExist):
            _delete_field_from_model(mutable_schema_info, alter_schema_instruction.field_name, version_change_name)
        elif isinstance(alter_schema_instruction, OldSchemaFieldHad):
            _change_field_in_model(
                mutable_schema_info,
                modified_schemas,
                alter_schema_instruction,
                version_change_name,
            )
        elif isinstance(alter_schema_instruction, OldSchemaFieldExistedWith):
            _add_field_to_model(mutable_schema_info, modified_schemas, alter_schema_instruction, version_change_name)
        elif isinstance(alter_schema_instruction, AlterSchemaInstruction):
            _change_model(mutable_schema_info, alter_schema_instruction, version_change_name)
        else:
            assert_never(alter_schema_instruction)


def _apply_alter_enum_instructions(
    enums: dict[IdentifierPythonPath, _EnumWrapper],
    alter_enum_instructions: Sequence[AlterEnumSubInstruction],
    version_change_name: str,
):
    for alter_enum_instruction in alter_enum_instructions:
        enum = alter_enum_instruction.enum
        enum_path = get_cls_pythonpath(enum)
        enum = enums[enum_path]
        if isinstance(alter_enum_instruction, EnumDidntHaveMembersInstruction):
            for member in alter_enum_instruction.members:
                if member not in enum.members:
                    raise InvalidGenerationInstructionError(
                        f'You tried to delete a member "{member}" from "{enum.cls.__name__}" '
                        f'in "{version_change_name}" but it doesn\'t have such a member.',
                    )
                enum.members.pop(member)
        elif isinstance(alter_enum_instruction, EnumHadMembersInstruction):
            for member, member_value in alter_enum_instruction.members.items():
                if member in enum.members and enum.members[member] == member_value:
                    raise InvalidGenerationInstructionError(
                        f'You tried to add a member "{member}" to "{enum.cls.__name__}" '
                        f'in "{version_change_name}" but there is already a member with that name and value.',
                    )
                enum.members[member] = member_value
        else:
            assert_never(alter_enum_instruction)


def _change_model(
    model: PydanticModelWrapper,
    alter_schema_instruction: AlterSchemaInstruction,
    version_change_name: str,
):
    # We only handle names right now so we just go ahead and check
    if alter_schema_instruction.name == model.name:
        raise InvalidGenerationInstructionError(
            f'You tried to change the name of "{model.name}" in "{version_change_name}" '
            "but it already has the name you tried to assign.",
        )
    model.name = alter_schema_instruction.name


def _add_field_to_model(
    model: PydanticModelWrapper,
    schemas: "dict[IdentifierPythonPath, PydanticModelWrapper]",
    alter_schema_instruction: OldSchemaFieldExistedWith,
    version_change_name: str,
):
    defined_fields = model._get_defined_fields(schemas)
    if alter_schema_instruction.field_name in defined_fields:
        raise InvalidGenerationInstructionError(
            f'You tried to add a field "{alter_schema_instruction.field_name}" to "{model.name}" '
            f'in "{version_change_name}" but there is already a field with that name.',
        )

    model.fields[alter_schema_instruction.field_name] = PydanticFieldWrapper(
        annotation_ast=None,
        annotation=alter_schema_instruction.type,
        init_model_field=alter_schema_instruction.field,
        field_ast=None,
    )


def _change_field_in_model(
    model: PydanticModelWrapper,
    schemas: "dict[IdentifierPythonPath, PydanticModelWrapper]",
    alter_schema_instruction: OldSchemaFieldHad,
    version_change_name: str,
):
    defined_fields = model._get_defined_fields(schemas)
    if alter_schema_instruction.field_name not in defined_fields:
        raise InvalidGenerationInstructionError(
            f'You tried to change the type of field "{alter_schema_instruction.field_name}" from '
            f'"{model.name}" in "{version_change_name}" but it doesn\'t have such a field.',
        )

    field = defined_fields[alter_schema_instruction.field_name]
    model.fields[alter_schema_instruction.field_name] = field

    current_field_is_constrained_type = is_pydantic_constrained_type(field.annotation)
    if alter_schema_instruction.type is not Sentinel:
        if field.annotation == alter_schema_instruction.type:
            raise InvalidGenerationInstructionError(
                f'You tried to change the type of field "{alter_schema_instruction.field_name}" to '
                f'"{alter_schema_instruction.type}" from "{model.name}" in "{version_change_name}" '
                f'but it already has type "{field.annotation}"',
            )
        field.annotation = alter_schema_instruction.type

        field.annotation_ast = None
        if current_field_is_constrained_type:
            field.field_ast = None

    if alter_schema_instruction.new_name is not Sentinel:
        if alter_schema_instruction.new_name == alter_schema_instruction.field_name:
            raise InvalidGenerationInstructionError(
                f'You tried to change the name of field "{alter_schema_instruction.field_name}" '
                f'from "{model.name}" in "{version_change_name}" '
                "but it already has that name.",
            )
        model.fields[alter_schema_instruction.new_name] = model.fields.pop(
            alter_schema_instruction.field_name,
        )

    field_info = field.field_info

    dict_of_field_info = {k: getattr(field_info, k) for k in field_info.__slots__}
    if dict_of_field_info == dict_of_empty_field_info:
        field_info = FieldInfo()
        field.field_info = field_info
    for attr_name in alter_schema_instruction.field_changes.__dataclass_fields__:
        attr_value = getattr(alter_schema_instruction.field_changes, attr_name)
        if attr_value is not Sentinel:
            if field.passed_field_attributes.get(attr_name, Sentinel) == attr_value:
                raise InvalidGenerationInstructionError(
                    f'You tried to change the attribute "{attr_name}" of field '
                    f'"{alter_schema_instruction.field_name}" '
                    f'from "{model.name}" to {attr_value!r} in "{version_change_name}" '
                    "but it already has that value.",
                )

            if hasattr(field.annotation, attr_name) and current_field_is_constrained_type:
                setattr(field.annotation, attr_name, attr_value)
                ann_ast = field.annotation_ast
                if ann_ast is not None and isinstance(ann_ast, ast.Call):
                    add_keyword_to_call(attr_name, attr_value, ann_ast)
                else:
                    field.field_ast = None
                    field.annotation_ast = None
            else:
                field.update_attribute(name=attr_name, value=attr_value)
                field_ast = field.field_ast
                if isinstance(field_ast, ast.Call):
                    add_keyword_to_call(attr_name, attr_value, field_ast)
                else:
                    field.field_ast = None


def _delete_field_from_model(model: PydanticModelWrapper, field_name: str, version_change_name: str):
    if field_name not in model.fields:
        raise InvalidGenerationInstructionError(
            f'You tried to delete a field "{field_name}" from "{model.name}" '
            f'in "{version_change_name}" but it doesn\'t have such a field.',
        )
    model.fields.pop(field_name)



import ast
import importlib
import os
import shutil
from collections.abc import Collection, Generator, Sequence
from copy import deepcopy
from pathlib import Path
from types import ModuleType
from typing import Any

import ast_comments

from cadwyn._package_utils import IdentifierPythonPath, get_package_path_from_module, get_version_dir_path
from cadwyn._utils import get_index_of_latest_schema_dir_in_module_python_path
from cadwyn.codegen._asts import get_all_names_defined_at_toplevel_of_module, read_python_module
from cadwyn.codegen._common import (
    CodegenContext,
    CodegenPlugin,
    GlobalCodegenContext,
    MigrationPlugin,
    PydanticModelWrapper,
    _EnumWrapper,
    _ModuleWrapper,
    get_fields_from_model,
)
from cadwyn.codegen._plugins.class_migrations import class_migration_plugin
from cadwyn.codegen._plugins.class_rebuilding import ClassRebuildingPlugin
from cadwyn.codegen._plugins.class_renaming import ClassRenamingPlugin
from cadwyn.codegen._plugins.import_auto_adding import ImportAutoAddingPlugin
from cadwyn.codegen._plugins.latest_version_aliasing import LatestVersionAliasingPlugin
from cadwyn.codegen._plugins.module_migrations import module_migration_plugin
from cadwyn.structure.versions import Version, VersionBundle

_AUTO_GENERATION_WARNING = "# THIS FILE WAS AUTO-GENERATED BY CADWYN. DO NOT EVER TRY TO EDIT IT BY HAND\n\n"


DEFAULT_CODEGEN_PLUGINS: tuple[CodegenPlugin, ...] = (
    LatestVersionAliasingPlugin(),
    ClassRebuildingPlugin(),
    ClassRenamingPlugin(),
    ImportAutoAddingPlugin(),
)
DEFAULT_CODEGEN_MIGRATION_PLUGINS: tuple[MigrationPlugin, ...] = (
    module_migration_plugin,
    class_migration_plugin,
)


def generate_code_for_versioned_packages(
    template_module: ModuleType,
    versions: VersionBundle,
    *,
    ignore_coverage_for_latest_aliases: bool = True,
    codegen_plugins: Sequence[CodegenPlugin] = DEFAULT_CODEGEN_PLUGINS,
    migration_plugins: Sequence[MigrationPlugin] = DEFAULT_CODEGEN_MIGRATION_PLUGINS,
    extra_context: dict[str, Any] | None = None,
):
    """
    Args:
        template_module: The latest package from which we will generate the versioned packages
        versions: Version bundle to generate versions from
        ignore_coverage_for_latest_aliases: Add a pragma: no cover comment to the star imports in the generated
        version of the latest module.
    """
    extra_context = extra_context or {}

    _generate_versioned_directories(
        template_module,
        versions=list(versions),
        schemas={
            k: PydanticModelWrapper(v, v.__name__, get_fields_from_model(v))
            for k, v in deepcopy(versions.versioned_schemas).items()
        },
        enums={
            k: _EnumWrapper(v, {member.name: member.value for member in v})
            for k, v in deepcopy(versions.versioned_enums).items()
        },
        modules={k: _ModuleWrapper(module) for k, module in versions.versioned_modules.items()},
        extra_context=extra_context | {"ignore_coverage_for_latest_aliases": ignore_coverage_for_latest_aliases},
        codegen_plugins=codegen_plugins,
        migration_plugins=migration_plugins,
    )


def _generate_versioned_directories(
    template_package: ModuleType,
    versions: list[Version],
    schemas: dict[IdentifierPythonPath, PydanticModelWrapper],
    enums: dict[IdentifierPythonPath, _EnumWrapper],
    modules: dict[IdentifierPythonPath, _ModuleWrapper],
    extra_context: dict[str, Any],
    codegen_plugins: Collection[CodegenPlugin],
    migration_plugins: Collection[MigrationPlugin],
):
    for version in versions:
        global_context = GlobalCodegenContext(
            current_version=version,
            versions=versions,
            schemas=schemas,
            enums=enums,
            modules=modules,
            extra=extra_context,
        )
        _generate_directory_for_version(template_package, codegen_plugins, version, global_context)
        for plugin in migration_plugins:
            plugin(global_context)


def _generate_directory_for_version(
    template_package: ModuleType,
    plugins: Collection[CodegenPlugin],
    version: Version,
    global_context: GlobalCodegenContext,
):
    template_dir = get_package_path_from_module(template_package)
    version_dir = get_version_dir_path(template_package, version.value)
    for (
        _relative_path_to_file,
        template_module,
        parallel_file,
    ) in _generate_parallel_directory(
        template_package,
        version_dir,
    ):
        file_source = read_python_module(template_module)
        parsed_file = ast.parse(file_source)
        context = _build_context(global_context, template_dir, version_dir, template_module, parallel_file, parsed_file)

        parsed_file = _apply_module_level_plugins(plugins, parsed_file, context)
        new_module = _apply_per_node_plugins(plugins, parsed_file, context)
        parallel_file.write_text(_AUTO_GENERATION_WARNING + ast_comments.unparse(new_module))


def _apply_module_level_plugins(
    plugins: Collection[CodegenPlugin],
    parsed_file: ast.Module,
    context: CodegenContext,
) -> ast.Module:
    node_type = type(parsed_file)
    for plugin in plugins:
        if issubclass(node_type, plugin.node_type):
            parsed_file = plugin(parsed_file, context)
    return parsed_file


def _apply_per_node_plugins(
    plugins: Collection[CodegenPlugin],
    parsed_file: ast.Module,
    context: CodegenContext,
) -> ast.Module:
    new_body = []

    for node in parsed_file.body:
        node_type = type(node)
        for plugin in plugins:
            if issubclass(node_type, plugin.node_type):
                node = plugin(node, context)  # noqa: PLW2901
        new_body.append(node)

    return ast.Module(body=new_body, type_ignores=[])


def _build_context(
    global_context: GlobalCodegenContext,
    template_dir: Path,
    version_dir: Path,
    template_module: ModuleType,
    parallel_file: Path,
    parsed_file: ast.Module,
):
    if template_module.__name__.endswith(".__init__"):
        module_python_path = template_module.__name__.removesuffix(".__init__")
    else:
        module_python_path = template_module.__name__
    all_names_defined_at_toplevel_of_file = get_all_names_defined_at_toplevel_of_module(
        parsed_file,
        module_python_path,
    )
    index_of_latest_package_dir_in_module_python_path = get_index_of_latest_schema_dir_in_module_python_path(
        template_module,
        template_dir.with_name(version_dir.name),
    )
    return CodegenContext(
        current_version=global_context.current_version,
        versions=global_context.versions,
        schemas=global_context.schemas,
        enums=global_context.enums,
        modules=global_context.modules,
        extra=global_context.extra,
        index_of_latest_package_dir_in_module_python_path=index_of_latest_package_dir_in_module_python_path,
        module_python_path=module_python_path,
        all_names_defined_on_toplevel_of_file=all_names_defined_at_toplevel_of_file,
        template_module=template_module,
        module_path=parallel_file,
    )


def _generate_parallel_directory(
    template_module: ModuleType,
    parallel_dir: Path,
) -> Generator[tuple[Path, ModuleType, Path], Any, None]:
    if template_module.__file__ is None:  # pragma: no cover
        raise ValueError(
            f"You passed a {template_module=} but it doesn't have a file "
            "so it is impossible to generate its counterpart.",
        )
    dir = get_package_path_from_module(template_module)
    parallel_dir.mkdir(exist_ok=True)
    # >>> [cadwyn, structure, schemas]
    template_module_python_path_parts = template_module.__name__.split(".")
    # >>> [home, foo, bar, cadwyn, structure, schemas]
    template_module_path_parts = Path(template_module.__file__).parent.parts
    # >>> [home, foo, bar] = [home, foo, bar, cadwyn, structure, schemas][:-3]
    root_module_path = Path(
        *template_module_path_parts[: -len(template_module_python_path_parts)],
    )
    for subroot, dirnames, filenames in os.walk(dir):
        original_subroot = Path(subroot)
        parallel_subroot = parallel_dir / original_subroot.relative_to(dir)
        if "__pycache__" in dirnames:
            dirnames.remove("__pycache__")
        for dirname in dirnames:
            (parallel_subroot / dirname).mkdir(exist_ok=True)
        for filename in filenames:
            original_file = (original_subroot / filename).absolute()
            parallel_file = (parallel_subroot / filename).absolute()

            if filename.endswith(".py"):
                original_module_path = ".".join(
                    original_file.relative_to(root_module_path).with_suffix("").parts,
                )
                original_module = importlib.import_module(original_module_path)
                yield original_subroot.relative_to(dir), original_module, parallel_file
            else:
                shutil.copyfile(original_file, parallel_file)



import ast
import inspect
from datetime import date
from pathlib import Path
from types import ModuleType

from cadwyn.exceptions import CodeGenerationError

IdentifierPythonPath = str


def get_absolute_python_path_of_import(node: ast.ImportFrom, module_python_path: str):
    python_path = ".".join(module_python_path.split(".")[0 : -node.level])
    result = []
    if node.module:
        result.append(node.module)
    if python_path:
        result.append(python_path)
    return ".".join(result)


def get_cls_pythonpath(cls: type) -> IdentifierPythonPath:
    return f"{cls.__module__}.{cls.__name__}"


def get_version_dir_path(template_package: ModuleType, version: date) -> Path:
    template_dir = get_package_path_from_module(template_package)
    return template_dir.with_name(get_version_dir_name(version))


def get_package_path_from_module(template_package: ModuleType) -> Path:
    # Can be cached in the future to gain some speedups
    file = inspect.getsourcefile(template_package)

    # I am too lazy to reproduce this error correctly
    if file is None:  # pragma: no cover
        raise CodeGenerationError(f'Package "{template_package}" has no source file')
    file = Path(file)
    if not file.name == "__init__.py":
        raise CodeGenerationError(f'"{template_package}" is not a package')
    return file.parent


def get_version_dir_name(version: date):
    return "v" + version.isoformat().replace("-", "_")



import email.message
import functools
import inspect
import json
from collections import defaultdict
from collections.abc import Callable, Sequence
from contextlib import AsyncExitStack
from contextvars import ContextVar
from enum import Enum
from types import ModuleType
from typing import Any, ClassVar, ParamSpec, TypeAlias, TypeVar, cast

from fastapi import HTTPException, params
from fastapi import Request as FastapiRequest
from fastapi import Response as FastapiResponse
from fastapi._compat import _normalize_errors
from fastapi.concurrency import run_in_threadpool
from fastapi.dependencies.models import Dependant
from fastapi.dependencies.utils import solve_dependencies
from fastapi.exceptions import RequestValidationError
from fastapi.responses import FileResponse, StreamingResponse
from fastapi.routing import APIRoute, _prepare_response_content
from pydantic import BaseModel
from starlette._utils import is_async_callable
from typing_extensions import assert_never

from cadwyn._compat import PYDANTIC_V2, ModelField, Undefined, model_dump
from cadwyn._package_utils import IdentifierPythonPath, get_cls_pythonpath
from cadwyn.exceptions import CadwynError, CadwynStructureError

from .._utils import Sentinel
from .common import Endpoint, VersionDate, VersionedModel
from .data import (
    AlterRequestByPathInstruction,
    AlterRequestBySchemaInstruction,
    AlterResponseByPathInstruction,
    AlterResponseBySchemaInstruction,
    RequestInfo,
    ResponseInfo,
)
from .endpoints import AlterEndpointSubInstruction
from .enums import AlterEnumSubInstruction
from .modules import AlterModuleInstruction
from .schemas import AlterSchemaInstruction, AlterSchemaSubInstruction

_CADWYN_REQUEST_PARAM_NAME = "cadwyn_request_param"
_CADWYN_RESPONSE_PARAM_NAME = "cadwyn_response_param"
_P = ParamSpec("_P")
_R = TypeVar("_R")
PossibleInstructions: TypeAlias = (
    AlterSchemaSubInstruction
    | AlterEndpointSubInstruction
    | AlterEnumSubInstruction
    | AlterSchemaInstruction
    | AlterModuleInstruction
    | staticmethod
)
APIVersionVarType: TypeAlias = ContextVar[VersionDate | None] | ContextVar[VersionDate]


class VersionChange:
    description: ClassVar[str] = Sentinel
    instructions_to_migrate_to_previous_version: ClassVar[Sequence[PossibleInstructions]] = Sentinel
    alter_schema_instructions: ClassVar[list[AlterSchemaSubInstruction | AlterSchemaInstruction]] = Sentinel
    alter_enum_instructions: ClassVar[list[AlterEnumSubInstruction]] = Sentinel
    alter_module_instructions: ClassVar[list[AlterModuleInstruction]] = Sentinel
    alter_endpoint_instructions: ClassVar[list[AlterEndpointSubInstruction]] = Sentinel
    alter_request_by_schema_instructions: ClassVar[dict[type[BaseModel], AlterRequestBySchemaInstruction]] = Sentinel
    alter_request_by_path_instructions: ClassVar[dict[str, list[AlterRequestByPathInstruction]]] = Sentinel
    alter_response_by_schema_instructions: ClassVar[dict[type, AlterResponseBySchemaInstruction]] = Sentinel
    alter_response_by_path_instructions: ClassVar[dict[str, list[AlterResponseByPathInstruction]]] = Sentinel
    _bound_version_bundle: "VersionBundle | None"

    def __init_subclass__(cls, _abstract: bool = False) -> None:
        super().__init_subclass__()

        if _abstract:
            return
        cls._validate_subclass()
        cls._extract_list_instructions_into_correct_containers()
        cls._extract_body_instructions_into_correct_containers()
        cls._check_no_subclassing()
        cls._bound_version_bundle = None

    @classmethod
    def _extract_body_instructions_into_correct_containers(cls):
        for instruction in cls.__dict__.values():
            if isinstance(instruction, AlterRequestBySchemaInstruction):
                if instruction.schema in cls.alter_request_by_schema_instructions:
                    raise CadwynStructureError(
                        f'There already exists a request migration for "{instruction.schema.__name__}" '
                        f'in "{cls.__name__}".',
                    )
                cls.alter_request_by_schema_instructions[instruction.schema] = instruction
            elif isinstance(instruction, AlterRequestByPathInstruction):
                cls.alter_request_by_path_instructions[instruction.path].append(instruction)
            elif isinstance(instruction, AlterResponseBySchemaInstruction):
                if instruction.schema in cls.alter_response_by_schema_instructions:
                    raise CadwynStructureError(
                        f'There already exists a response migration for "{instruction.schema.__name__}" '
                        f'in "{cls.__name__}".',
                    )
                cls.alter_response_by_schema_instructions[instruction.schema] = instruction
            elif isinstance(instruction, AlterResponseByPathInstruction):
                cls.alter_response_by_path_instructions[instruction.path].append(instruction)

    @classmethod
    def _extract_list_instructions_into_correct_containers(cls):
        cls.alter_schema_instructions = []
        cls.alter_enum_instructions = []
        cls.alter_module_instructions = []
        cls.alter_endpoint_instructions = []
        cls.alter_request_by_schema_instructions = {}
        cls.alter_request_by_path_instructions = defaultdict(list)
        cls.alter_response_by_schema_instructions = {}
        cls.alter_response_by_path_instructions = defaultdict(list)
        for alter_instruction in cls.instructions_to_migrate_to_previous_version:
            if isinstance(alter_instruction, AlterSchemaInstruction | AlterSchemaSubInstruction):
                cls.alter_schema_instructions.append(alter_instruction)
            elif isinstance(alter_instruction, AlterEnumSubInstruction):
                cls.alter_enum_instructions.append(alter_instruction)
            elif isinstance(alter_instruction, AlterModuleInstruction):
                cls.alter_module_instructions.append(alter_instruction)
            elif isinstance(alter_instruction, AlterEndpointSubInstruction):
                cls.alter_endpoint_instructions.append(alter_instruction)
            elif isinstance(alter_instruction, staticmethod):  # pragma: no cover
                raise NotImplementedError(f'"{alter_instruction}" is an unacceptable version change instruction')
            else:
                assert_never(alter_instruction)

    @classmethod
    def _validate_subclass(cls):
        if cls.description is Sentinel:
            raise CadwynStructureError(
                f"Version change description is not set on '{cls.__name__}' but is required.",
            )
        if cls.instructions_to_migrate_to_previous_version is Sentinel:
            raise CadwynStructureError(
                f"Attribute 'instructions_to_migrate_to_previous_version' is not set on '{cls.__name__}'"
                " but is required.",
            )
        if not isinstance(cls.instructions_to_migrate_to_previous_version, Sequence):
            raise CadwynStructureError(
                f"Attribute 'instructions_to_migrate_to_previous_version' must be a sequence in '{cls.__name__}'.",
            )
        for instruction in cls.instructions_to_migrate_to_previous_version:
            if not isinstance(instruction, PossibleInstructions):
                raise CadwynStructureError(
                    f"Instruction '{instruction}' is not allowed. Please, use the correct instruction types",
                )
        for attr_name, attr_value in cls.__dict__.items():
            if not isinstance(
                attr_value,
                AlterRequestBySchemaInstruction
                | AlterRequestByPathInstruction
                | AlterResponseBySchemaInstruction
                | AlterResponseByPathInstruction,
            ) and attr_name not in {
                "description",
                "side_effects",
                "instructions_to_migrate_to_previous_version",
                "__module__",
                "__doc__",
            }:
                raise CadwynStructureError(
                    f"Found: '{attr_name}' attribute of type '{type(attr_value)}' in '{cls.__name__}'."
                    " Only migration instructions and schema properties are allowed in version change class body.",
                )

    @classmethod
    def _check_no_subclassing(cls):
        if cls.mro() != [cls, VersionChange, object]:
            raise TypeError(
                f"Can't subclass {cls.__name__} as it was never meant to be subclassed.",
            )

    def __init__(self) -> None:  # pyright: ignore[reportMissingSuperCall]
        raise TypeError(
            f"Can't instantiate {self.__class__.__name__} as it was never meant to be instantiated.",
        )


class VersionChangeWithSideEffects(VersionChange, _abstract=True):
    @classmethod
    def _check_no_subclassing(cls):
        if cls.mro() != [cls, VersionChangeWithSideEffects, VersionChange, object]:
            raise TypeError(
                f"Can't subclass {cls.__name__} as it was never meant to be subclassed.",
            )

    @classmethod
    @property
    def is_applied(cls) -> bool:
        if (
            cls._bound_version_bundle is None
            or cls not in cls._bound_version_bundle._version_changes_to_version_mapping
        ):
            raise CadwynError(
                f"You tried to check whether '{cls.__name__}' is active but it was never bound to any version.",
            )
        api_version = cls._bound_version_bundle.api_version_var.get()
        if api_version is None:
            return True
        return cls._bound_version_bundle._version_changes_to_version_mapping[cls] <= api_version


class Version:
    def __init__(self, value: VersionDate, *version_changes: type[VersionChange]) -> None:
        super().__init__()

        self.value = value
        self.version_changes = version_changes

    def __repr__(self) -> str:
        return f"Version('{self.value}')"


class VersionBundle:
    def __init__(
        self,
        latest_version: Version,
        /,
        *other_versions: Version,
        api_version_var: APIVersionVarType | None = None,
    ) -> None:
        super().__init__()

        self.versions = (latest_version, *other_versions)
        if api_version_var is None:
            api_version_var = ContextVar("cadwyn_api_version")
        self.api_version_var = api_version_var
        if sorted(self.versions, key=lambda v: v.value, reverse=True) != list(self.versions):
            raise CadwynStructureError(
                "Versions are not sorted correctly. Please sort them in descending order.",
            )
        if self.versions[-1].version_changes:
            raise CadwynStructureError(
                f'The first version "{self.versions[-1].value}" cannot have any version changes. '
                "Version changes are defined to migrate to/from a previous version so you "
                "cannot define one for the very first version.",
            )
        version_values = set()
        for version in self.versions:
            if version.value not in version_values:
                version_values.add(version.value)
            else:
                raise CadwynStructureError(
                    f"You tried to define two versions with the same value in the same "
                    f"{VersionBundle.__name__}: '{version.value}'.",
                )
            for version_change in version.version_changes:
                if version_change._bound_version_bundle is not None:
                    raise CadwynStructureError(
                        f"You tried to bind version change '{version_change.__name__}' to two different versions. "
                        "It is prohibited.",
                    )
                version_change._bound_version_bundle = self

    def __iter__(self):
        yield from self.versions

    @functools.cached_property
    def versioned_schemas(self) -> dict[IdentifierPythonPath, type[VersionedModel]]:
        return {
            get_cls_pythonpath(instruction.schema): instruction.schema
            for version in self.versions
            for version_change in version.version_changes
            for instruction in list(version_change.alter_schema_instructions)
            + list(version_change.alter_request_by_schema_instructions.values())
        }

    @functools.cached_property
    def versioned_enums(self) -> dict[IdentifierPythonPath, type[Enum]]:
        return {
            get_cls_pythonpath(instruction.enum): instruction.enum
            for version in self.versions
            for version_change in version.version_changes
            for instruction in version_change.alter_enum_instructions
        }

    @functools.cached_property
    def versioned_modules(self) -> dict[IdentifierPythonPath, ModuleType]:
        return {
            # We do this because when users import their modules, they might import
            # the __init__.py file directly instead of the package itself
            # which results in this extra `.__init__` suffix in the name
            instruction.module.__name__.removesuffix(".__init__"): instruction.module
            for version in self.versions
            for version_change in version.version_changes
            for instruction in version_change.alter_module_instructions
        }

    @functools.cached_property
    def _version_changes_to_version_mapping(
        self,
    ) -> dict[type[VersionChange], VersionDate]:
        return {
            version_change: version.value for version in self.versions for version_change in version.version_changes
        }

    async def _migrate_request(
        self,
        body_type: type[BaseModel] | None,
        latest_dependant_with_internal_schema: Dependant,
        request: FastapiRequest,
        response: FastapiResponse,
        request_info: RequestInfo,
        current_version: VersionDate,
        latest_route: APIRoute,
    ) -> dict[str, Any]:
        path = request.scope["path"]
        method = request.method
        for v in reversed(self.versions):
            if v.value <= current_version:
                continue
            for version_change in v.version_changes:
                if body_type is not None and body_type in version_change.alter_request_by_schema_instructions:
                    version_change.alter_request_by_schema_instructions[body_type](request_info)
                if path in version_change.alter_request_by_path_instructions:
                    for instruction in version_change.alter_request_by_path_instructions[path]:
                        if method in instruction.methods:
                            instruction(request_info)
        request.scope["headers"] = tuple((key.encode(), value.encode()) for key, value in request_info.headers.items())
        del request._headers
        # Remember this: if len(body_params) == 1, then route.body_schema == route.dependant.body_params[0]
        async with AsyncExitStack() as async_exit_stack:
            new_kwargs, errors, _, _, _ = await solve_dependencies(
                request=request,
                response=response,
                dependant=latest_dependant_with_internal_schema,
                body=request_info.body,
                dependency_overrides_provider=latest_route.dependency_overrides_provider,
                async_exit_stack=async_exit_stack,
            )
            if errors:
                raise RequestValidationError(_normalize_errors(errors), body=request_info.body)
            return new_kwargs
        raise NotImplementedError("This code should not be reachable. If it was reached -- it's a bug.")

    def _migrate_response(
        self,
        response_info: ResponseInfo,
        current_version: VersionDate,
        latest_route: APIRoute,
        path: str,
        method: str,
    ) -> ResponseInfo:
        """Convert the data to a specific version by applying all version changes in reverse order.

        Args:
            endpoint: the function which usually returns this data. Data migrations marked with this endpoint will
            be applied to the passed data
            payload: data to be migrated. Will be mutated during the call
            version: the version to which the data should be converted

        Returns:
            Modified data
        """

        for v in self.versions:
            if v.value <= current_version:
                break
            for version_change in v.version_changes:
                if (
                    latest_route.response_model
                    and latest_route.response_model in version_change.alter_response_by_schema_instructions
                ):
                    version_change.alter_response_by_schema_instructions[latest_route.response_model](response_info)
                if path in version_change.alter_response_by_path_instructions:
                    for instruction in version_change.alter_response_by_path_instructions[path]:
                        if method in instruction.methods:
                            instruction(response_info)
        return response_info

    # TODO (https://github.com/zmievsa/cadwyn/issues/113): Refactor this function and all functions it calls.
    def _versioned(
        self,
        template_module_body_field_for_request_migrations: type[BaseModel] | None,
        module_body_field_name: str | None,
        route: APIRoute,
        latest_route: APIRoute,
        dependant_for_request_migrations: Dependant,
        *,
        request_param_name: str,
        response_param_name: str,
    ) -> Callable[[Endpoint[_P, _R]], Endpoint[_P, _R]]:
        def wrapper(endpoint: Endpoint[_P, _R]) -> Endpoint[_P, _R]:
            @functools.wraps(endpoint)
            async def decorator(*args: Any, **kwargs: Any) -> _R:
                request: FastapiRequest = kwargs[request_param_name]
                response: FastapiResponse = kwargs[response_param_name]
                path = request.scope["path"]
                method = request.method
                kwargs = await self._convert_endpoint_kwargs_to_version(
                    template_module_body_field_for_request_migrations,
                    module_body_field_name,
                    # Dependant must be from the version of the finally migrated request, not the version of endpoint
                    dependant_for_request_migrations,
                    request_param_name,
                    kwargs,
                    response,
                    route,
                    latest_route,
                )

                return await self._convert_endpoint_response_to_version(
                    endpoint,
                    latest_route,
                    path,
                    method,
                    response_param_name,
                    kwargs,
                    response,
                )

            if request_param_name == _CADWYN_REQUEST_PARAM_NAME:
                _add_keyword_only_parameter(decorator, _CADWYN_REQUEST_PARAM_NAME, FastapiRequest)
            if response_param_name == _CADWYN_RESPONSE_PARAM_NAME:
                _add_keyword_only_parameter(decorator, _CADWYN_RESPONSE_PARAM_NAME, FastapiResponse)

            return decorator  # pyright: ignore[reportGeneralTypeIssues]

        return wrapper

    async def _convert_endpoint_response_to_version(
        self,
        func_to_get_response_from: Endpoint,
        latest_route: APIRoute,
        path: str,
        method: str,
        response_param_name: str,
        kwargs: dict[str, Any],
        fastapi_response_dependency: FastapiResponse,
    ) -> Any:
        if response_param_name == _CADWYN_RESPONSE_PARAM_NAME:
            kwargs.pop(response_param_name)
        if is_async_callable(func_to_get_response_from):
            response_or_response_body: FastapiResponse | object = await func_to_get_response_from(**kwargs)
        else:
            response_or_response_body: FastapiResponse | object = await run_in_threadpool(
                func_to_get_response_from,
                **kwargs,
            )
        api_version = self.api_version_var.get()
        if api_version is None:
            return response_or_response_body

        if isinstance(response_or_response_body, FastapiResponse):
            # TODO (https://github.com/zmievsa/cadwyn/issues/125): Add support for migrating `StreamingResponse`
            # TODO (https://github.com/zmievsa/cadwyn/issues/126): Add support for migrating `FileResponse`
            # Starlette breaks Liskov Substitution principle and
            # doesn't define `body` for `StreamingResponse` and `FileResponse`
            if isinstance(response_or_response_body, StreamingResponse | FileResponse):
                body = None
            elif response_or_response_body.body:
                body = json.loads(response_or_response_body.body)
            else:
                body = None
                # TODO (https://github.com/zmievsa/cadwyn/issues/51): Only do this if there are migrations
            response_info = ResponseInfo(response_or_response_body, body)
        else:
            response_info = ResponseInfo(
                fastapi_response_dependency,
                _prepare_response_content(
                    response_or_response_body,
                    exclude_unset=latest_route.response_model_exclude_unset,
                    exclude_defaults=latest_route.response_model_exclude_defaults,
                    exclude_none=latest_route.response_model_exclude_none,
                ),
            )

        response_info = self._migrate_response(
            response_info,
            api_version,
            latest_route,
            path,
            method,
        )
        if isinstance(response_or_response_body, FastapiResponse):
            # a webserver (uvicorn for instance) calculates the body at the endpoint level.
            # if an endpoint returns no "body", its content-length will be set to 0
            # json.dumps(None) results into "null", and content-length should be 4,
            # but it was already calculated to 0 which causes
            # `RuntimeError: Response content longer than Content-Length` or
            # `Too much data for declared Content-Length`, based on the protocol
            # which is why we skip the None case.

            # We skip cases without "body" attribute because of StreamingResponse and FileResponse
            # that do not have it. We don't support it too.
            if response_info.body is not None and hasattr(response_info._response, "body"):
                # TODO (https://github.com/zmievsa/cadwyn/issues/51): Only do this if there are migrations
                response_info._response.body = json.dumps(response_info.body).encode()
            return response_info._response
        return response_info.body

    async def _convert_endpoint_kwargs_to_version(
        self,
        template_module_body_field_for_request_migrations: type[BaseModel] | None,
        body_field_alias: str | None,
        latest_dependant_with_internal_schema: Dependant,
        request_param_name: str,
        kwargs: dict[str, Any],
        response: FastapiResponse,
        route: APIRoute,
        latest_route: APIRoute,
    ) -> dict[str, Any]:
        request: FastapiRequest = kwargs[request_param_name]
        if request_param_name == _CADWYN_REQUEST_PARAM_NAME:
            kwargs.pop(request_param_name)

        api_version = self.api_version_var.get()
        if api_version is None:
            return kwargs

        # This is a kind of body param you get when you define a single pydantic schema in your route's body
        if (
            len(route.dependant.body_params) == 1
            and template_module_body_field_for_request_migrations is not None
            and body_field_alias is not None
            and body_field_alias in kwargs
        ):
            raw_body: BaseModel | None = kwargs.get(body_field_alias)
            if raw_body is None:
                body = None
            else:
                body = model_dump(raw_body, by_alias=True, exclude_unset=True)
                if not PYDANTIC_V2 and raw_body.__custom_root_type__:  # pyright: ignore[reportGeneralTypeIssues]
                    body = body["__root__"]
        else:
            # This is for requests without body or with complex body such as form or file
            body = await _get_body(request, route.body_field)

        request_info = RequestInfo(request, body)
        new_kwargs = await self._migrate_request(
            template_module_body_field_for_request_migrations,
            latest_dependant_with_internal_schema,
            request,
            response,
            request_info,
            api_version,
            latest_route,
        )
        # Because we re-added it into our kwargs when we did solve_dependencies
        if _CADWYN_REQUEST_PARAM_NAME in new_kwargs:
            new_kwargs.pop(_CADWYN_REQUEST_PARAM_NAME)

        return new_kwargs


# We use this instead of `.body()` to automatically guess body type and load the correct body, even if it's a form
async def _get_body(request: FastapiRequest, body_field: ModelField | None):  # pragma: no cover # This is from fastapi
    is_body_form = body_field and isinstance(body_field.field_info, params.Form)
    try:
        body: Any = None
        if body_field:
            if is_body_form:
                body = await request.form()
                stack = cast(AsyncExitStack, request.scope.get("fastapi_astack"))
                stack.push_async_callback(body.close)
            else:
                body_bytes = await request.body()
                if body_bytes:
                    json_body: Any = Undefined
                    content_type_value = request.headers.get("content-type")
                    if not content_type_value:
                        json_body = await request.json()
                    else:
                        message = email.message.Message()
                        message["content-type"] = content_type_value
                        if message.get_content_maintype() == "application":
                            subtype = message.get_content_subtype()
                            if subtype == "json" or subtype.endswith("+json"):
                                json_body = await request.json()
                    if json_body != Undefined:
                        body = json_body
                    else:
                        body = body_bytes
    except json.JSONDecodeError as e:
        raise RequestValidationError(
            [
                {
                    "type": "json_invalid",
                    "loc": ("body", e.pos),
                    "msg": "JSON decode error",
                    "input": {},
                    "ctx": {"error": e.msg},
                },
            ],
            body=e.doc,
        ) from e
    except HTTPException:
        raise
    except Exception as e:  # noqa: BLE001
        raise HTTPException(status_code=400, detail="There was an error parsing the body") from e
    return body


def _add_keyword_only_parameter(
    func: Callable,
    param_name: str,
    param_annotation: type,
):
    signature = inspect.signature(func)
    func.__signature__ = signature.replace(
        parameters=(
            [
                *list(signature.parameters.values()),
                inspect.Parameter(param_name, kind=inspect._ParameterKind.KEYWORD_ONLY, annotation=param_annotation),
            ]
        ),
    )



import ast
import copy
from typing import Any

from cadwyn._compat import (
    PydanticFieldWrapper,
    get_attrs_that_are_not_from_field_and_that_are_from_field,
    is_pydantic_constrained_type,
)
from cadwyn._package_utils import IdentifierPythonPath, get_absolute_python_path_of_import
from cadwyn.codegen._asts import (
    get_ast_keyword_from_argument_name_and_value,
    get_fancy_repr,
    pop_docstring_from_cls_body,
)
from cadwyn.codegen._common import CodegenContext, PydanticModelWrapper, _EnumWrapper


class ClassRebuildingPlugin:
    node_type = ast.Module

    @staticmethod
    def __call__(node: ast.Module, context: CodegenContext) -> Any:
        if context.current_version_is_latest:
            return node

        node.body = [_migrate_ast_node_to_another_version(n, context) for n in node.body]

        return node


def _migrate_ast_node_to_another_version(
    node: ast.stmt,
    context: CodegenContext,
):
    if isinstance(node, ast.ClassDef):
        return _migrate_cls_to_another_version(node, context)
    elif isinstance(node, ast.ImportFrom):
        python_path = get_absolute_python_path_of_import(node, context.module_python_path)
        node.names = [
            name
            if (name_path := f"{python_path}.{name.name}") not in context.schemas
            else ast.alias(name=context.schemas[name_path].name, asname=name.asname)
            for name in node.names
        ]

    return node


def _migrate_cls_to_another_version(
    cls_node: ast.ClassDef,
    context: CodegenContext,
) -> ast.ClassDef:
    cls_python_path = f"{context.module_python_path}.{cls_node.name}"

    if cls_python_path in context.schemas:
        cls_node = _modify_schema_cls(cls_node, context.schemas, cls_python_path)
    elif cls_python_path in context.enums:
        cls_node = _modify_enum_cls(cls_node, context.enums[cls_python_path])

    if not cls_node.body:
        cls_node.body = [ast.Pass()]
    return cls_node


def _modify_enum_cls(cls_node: ast.ClassDef, enum: _EnumWrapper) -> ast.ClassDef:
    new_body = [
        ast.Assign(
            targets=[ast.Name(member, ctx=ast.Store())],
            value=ast.Name(get_fancy_repr(member_value)),
            lineno=0,
        )
        for member, member_value in enum.members.items()
    ]

    old_body = [n for n in cls_node.body if not isinstance(n, ast.AnnAssign | ast.Assign | ast.Pass | ast.Constant)]
    docstring = pop_docstring_from_cls_body(old_body)

    cls_node.body = docstring + new_body + old_body
    return cls_node


def _modify_schema_cls(
    cls_node: ast.ClassDef,
    modified_schemas: dict[IdentifierPythonPath, PydanticModelWrapper],
    cls_python_path: str,
) -> ast.ClassDef:
    model_info = modified_schemas[cls_python_path]
    # This is for possible schema renaming
    cls_node.name = model_info.name

    field_definitions = [
        ast.AnnAssign(
            target=ast.Name(name, ctx=ast.Store()),
            annotation=_render_annotation(field.get_annotation_for_rendering()),
            value=_generate_field_ast(field),
            simple=1,
        )
        for name, field in model_info.fields.items()
    ]

    old_body = [n for n in cls_node.body if not isinstance(n, ast.AnnAssign | ast.Assign | ast.Pass | ast.Constant)]
    docstring = pop_docstring_from_cls_body(old_body)
    cls_node.body = docstring + field_definitions + old_body
    if not cls_node.body:
        cls_node.body = [ast.Pass()]
    return cls_node


def _render_annotation(annotation: Any):
    if isinstance(annotation, ast.AST):
        return copy.deepcopy(annotation)
    return ast.parse(get_fancy_repr(annotation), mode="eval").body


def _generate_field_ast(field: PydanticFieldWrapper):
    if field.field_ast is not None:
        # We do this because next plugins **might** use a transformer which will edit the ast within the field
        # and break rendering
        return copy.deepcopy(field.field_ast)
    passed_attrs = field.passed_field_attributes
    if is_pydantic_constrained_type(field.annotation) and field.annotation_ast is None:
        (
            attrs_that_are_only_in_contype,
            attrs_that_are_only_in_field,
        ) = get_attrs_that_are_not_from_field_and_that_are_from_field(field.annotation)
        if not attrs_that_are_only_in_contype:
            passed_attrs |= attrs_that_are_only_in_field

    if passed_attrs:
        return ast.Call(
            func=ast.Name("Field"),
            args=[],
            keywords=[
                get_ast_keyword_from_argument_name_and_value(attr, attr_value)
                for attr, attr_value in passed_attrs.items()
            ],
        )
    return None



import ast

from cadwyn._compat import PYDANTIC_V2
from cadwyn.codegen._common import CodegenContext

_extra_imports: list[tuple[str, str]] = [
    ("typing", "import typing"),
    ("Any", "from typing import Any"),
    ("Annotated", "from typing import Annotated"),
    ("Field", "from pydantic import Field"),
    ("conbytes", "from pydantic import conbytes"),
    ("conlist", "from pydantic import conlist"),
    ("conset", "from pydantic import conset"),
    ("constr", "from pydantic import constr"),
    ("conint", "from pydantic import conint"),
    ("confloat", "from pydantic import confloat"),
    ("condecimal", "from pydantic import condecimal"),
    ("condate", "from pydantic import condate"),
]


if PYDANTIC_V2:
    _extra_imports.extend(
        [
            ("confrozenset", "from pydantic import conset"),
            ("StringConstraints", "from pydantic import StringConstraints"),
            ("StrictBool", "from pydantic import StrictBool"),
            ("StrictBytes", "from pydantic import StrictBytes"),
            ("StrictFloat", "from pydantic import StrictFloat"),
            ("StrictInt", "from pydantic import StrictInt"),
            ("StrictStr", "from pydantic import StrictStr"),
        ],
    )

_rendered_extra_imports = [(seek_str, ast.parse(imp).body[0]) for seek_str, imp in _extra_imports]


class ImportAutoAddingPlugin:
    node_type = ast.Module

    @staticmethod
    def __call__(node: ast.Module, context: CodegenContext):
        if context.current_version_is_latest:
            return node
        source = ast.unparse(node)
        extra_lib_imports = [
            import_
            for seek_str, import_ in _rendered_extra_imports
            if seek_str in source and seek_str not in context.all_names_defined_on_toplevel_of_file
        ]
        # We do this because when we import our module, we import not the package but
        # the __init__.py file directly which produces this extra `.__init__` suffix in the name
        module_name = context.template_module.__name__.removesuffix(".__init__")
        if module_name in context.modules:
            manual_imports = context.modules[module_name].extra_imports
        else:
            manual_imports = []

        node.body = extra_lib_imports + manual_imports + node.body

        return node



import ast
from typing import Any

from cadwyn._package_utils import IdentifierPythonPath
from cadwyn.codegen._common import CodegenContext, PydanticModelWrapper


class ClassRenamingPlugin:
    node_type = ast.Module

    @staticmethod
    def __call__(
        node: ast.Module,
        context: CodegenContext,
    ):
        if context.current_version_is_latest:
            return node
        return _AnnotationASTNodeTransformerWithSchemaRenaming(
            context.schemas,
            context.all_names_defined_on_toplevel_of_file,
            context.module_python_path,
        ).visit(node)


class _AnnotationASTNodeTransformerWithSchemaRenaming(ast.NodeTransformer):
    def __init__(
        self,
        modified_schemas: dict[IdentifierPythonPath, PydanticModelWrapper],
        all_names_in_module: dict[str, str],
        module_python_path: str,
    ):
        super().__init__()
        self.modified_schemas = modified_schemas
        self.module_python_path = module_python_path
        self.all_names_in_module = all_names_in_module

    def visit_AnnAssign(self, node: ast.AnnAssign):  # noqa: N802
        # Handles Pydantic annotations that are strings
        if isinstance(node.annotation, ast.Constant) and isinstance(node.annotation.value, str):
            altered = self.visit(ast.parse(node.annotation.value, mode="eval").body)
            node.annotation.value = ast.unparse(altered)
        return self.generic_visit(node)

    def visit_Name(self, node: ast.Name) -> Any:  # noqa: N802
        return self._get_name(node, node.id)

    def _get_name(self, node: ast.AST, name: str):
        model_info = self.modified_schemas.get(f"{self.all_names_in_module.get(name, self.module_python_path)}.{name}")
        if model_info is not None:
            return ast.Name(model_info.name)
        return node



from typing_extensions import assert_never

from cadwyn.codegen._common import GlobalCodegenContext
from cadwyn.structure.modules import AlterModuleInstruction


def module_migration_plugin(context: GlobalCodegenContext):
    for version_change in context.current_version.version_changes:
        for alter_module_instruction in version_change.alter_module_instructions:
            module = alter_module_instruction.module
            mutable_module = context.modules[module.__name__]
            if isinstance(alter_module_instruction, AlterModuleInstruction):
                mutable_module.extra_imports.append(alter_module_instruction.import_)
            else:
                assert_never(alter_module_instruction)



import ast
import dataclasses
import inspect
from typing import Any, TypeAlias

from fastapi._compat import ModelField as FastAPIModelField
from pydantic import BaseModel, Field

ModelField: TypeAlias = Any  # pyright: ignore[reportGeneralTypeIssues]
Undefined: TypeAlias = Any

try:
    PYDANTIC_V2 = False

    from pydantic.fields import FieldInfo, ModelField, Undefined  # pyright: ignore # noqa: PGH003

    _all_field_arg_names = []
    EXTRA_FIELD_NAME = "extra"
except ImportError:
    PYDANTIC_V2 = True

    from pydantic.fields import FieldInfo
    from pydantic_core import PydanticUndefined

    ModelField: TypeAlias = FieldInfo  # pyright: ignore # noqa: PGH003
    Undefined = PydanticUndefined  # pyright: ignore # noqa: PGH003
    _all_field_arg_names = sorted(
        [
            name
            for name, param in inspect.signature(Field).parameters.items()
            if param.kind in {inspect._ParameterKind.KEYWORD_ONLY, inspect._ParameterKind.POSITIONAL_OR_KEYWORD}
        ],
    )
    EXTRA_FIELD_NAME = "json_schema_extra"


_empty_field_info = Field()
dict_of_empty_field_info = {k: getattr(_empty_field_info, k) for k in FieldInfo.__slots__}


def is_pydantic_constrained_type(value: object):
    return isinstance(value, type) and value.__name__.startswith("Constrained") and value.__name__.endswith("Value")


def get_attrs_that_are_not_from_field_and_that_are_from_field(value: type):
    parent_public_attrs = {k: v for k, v in value.mro()[1].__dict__.items() if not k.startswith("_")}
    value_private_attrs = {k: v for k, v in value.__dict__.items() if not k.startswith("_")}
    attrs_in_value_different_from_parent = {
        k: v for k, v in value_private_attrs.items() if k in parent_public_attrs and parent_public_attrs[k] != v
    }
    attrs_in_value_different_from_parent_that_are_not_in_field_def = {
        k: v for k, v in attrs_in_value_different_from_parent.items() if k not in dict_of_empty_field_info
    }
    attrs_in_value_different_from_parent_that_are_in_field_def = {
        k: v for k, v in attrs_in_value_different_from_parent.items() if k in dict_of_empty_field_info
    }

    return (
        attrs_in_value_different_from_parent_that_are_not_in_field_def,
        attrs_in_value_different_from_parent_that_are_in_field_def,
    )


@dataclasses.dataclass(slots=True)
class PydanticFieldWrapper:
    annotation: Any

    init_model_field: dataclasses.InitVar[ModelField]  # pyright: ignore[reportGeneralTypeIssues]
    field_info: FieldInfo = dataclasses.field(init=False)

    annotation_ast: ast.expr | None = None
    field_ast: ast.expr | None = None

    def __post_init__(self, init_model_field: ModelField):  # pyright: ignore[reportGeneralTypeIssues]
        if isinstance(init_model_field, FieldInfo):
            self.field_info = init_model_field
        else:
            self.field_info = init_model_field.field_info

    def get_annotation_for_rendering(self):
        if self.annotation_ast:
            return self.annotation_ast
        else:
            return self.annotation

    def update_attribute(self, *, name: str, value: Any):
        if PYDANTIC_V2:
            if name in FieldInfo.metadata_lookup:
                self.field_info.metadata.extend(FieldInfo._collect_metadata({name: value}))
            self.field_info._attributes_set[name] = value
        else:
            setattr(self.field_info, name, value)

    @property
    def passed_field_attributes(self):
        if PYDANTIC_V2:
            attributes = {
                attr_name: self.field_info._attributes_set[attr_name]
                for attr_name in _all_field_arg_names
                if attr_name in self.field_info._attributes_set
            }
            # PydanticV2 always adds frozen to _attributes_set but we don't want it if it wasn't explicitly set
            if attributes.get("frozen", ...) is None:
                attributes.pop("frozen")
            return attributes

        else:
            attributes = {
                attr_name: attr_val
                for attr_name, default_attr_val in dict_of_empty_field_info.items()
                if attr_name != EXTRA_FIELD_NAME
                and (attr_val := getattr(self.field_info, attr_name)) != default_attr_val
            }
            extras = getattr(self.field_info, EXTRA_FIELD_NAME) or {}
            return attributes | extras


def model_fields(model: type[BaseModel]) -> dict[str, FieldInfo]:
    if PYDANTIC_V2:
        return model.model_fields
    else:
        return model.__fields__  # pyright: ignore[reportDeprecated]


def model_dump(model: BaseModel, by_alias: bool = False, exclude_unset: bool = False) -> dict[str, Any]:
    if PYDANTIC_V2:
        return model.model_dump(by_alias=by_alias, exclude_unset=exclude_unset)
    else:
        return model.dict(by_alias=by_alias, exclude_unset=exclude_unset)  # pyright: ignore[reportDeprecated]


def rebuild_fastapi_body_param(old_body_param: FastAPIModelField, new_body_param_type: type[BaseModel]):
    kwargs: dict[str, Any] = {"name": old_body_param.name, "field_info": old_body_param.field_info}
    if PYDANTIC_V2:
        old_body_param.field_info.annotation = new_body_param_type
        kwargs.update({"mode": old_body_param.mode})
    else:
        kwargs.update(
            {
                "type_": new_body_param_type,
                "class_validators": old_body_param.class_validators,  # pyright: ignore[reportGeneralTypeIssues]
                "default": old_body_param.default,
                "required": old_body_param.required,
                "model_config": old_body_param.model_config,  # pyright: ignore[reportGeneralTypeIssues]
                "alias": old_body_param.alias,
            },
        )
    return FastAPIModelField(**kwargs)



from .data import (
    RequestInfo,
    ResponseInfo,
    convert_request_to_next_version_for,
    convert_response_to_previous_version_for,
)
from .endpoints import endpoint
from .enums import enum
from .modules import module
from .schemas import schema
from .versions import (
    Version,
    VersionBundle,
    VersionChange,
    VersionChangeWithSideEffects,
)

__all__ = [
    "VersionBundle",
    "Version",
    "VersionChange",
    "VersionChangeWithSideEffects",
    "endpoint",
    "schema",
    "enum",
    "module",
    "convert_response_to_previous_version_for",
    "convert_request_to_next_version_for",
    "RequestInfo",
    "ResponseInfo",
]



import ast
import dataclasses
from dataclasses import InitVar, dataclass
from types import ModuleType

from cadwyn.exceptions import CadwynStructureError


@dataclass(slots=True)
class AlterModuleInstruction:
    module: ModuleType
    raw_import: InitVar[str]
    import_: ast.Import | ast.ImportFrom = dataclasses.field(init=False)

    def __post_init__(self, raw_import: str):
        parsed_body = ast.parse(raw_import).body
        if len(parsed_body) > 1:
            raise CadwynStructureError(
                f"You have specified more than just a single import. This is prohibited. "
                f"Problematic string: {raw_import}",
            )
        if not isinstance(parsed_body[0], ast.Import | ast.ImportFrom):
            raise CadwynStructureError(
                f"You have specified a non-import statement. This is prohibited. Problematic string: {raw_import}",
            )
        self.import_ = parsed_body[0]


@dataclass(slots=True)
class AlterModuleInstructionFactory:
    module: ModuleType

    def had(self, *, import_: str) -> AlterModuleInstruction:
        return AlterModuleInstruction(self.module, import_)


def module(module: ModuleType, /) -> AlterModuleInstructionFactory:
    return AlterModuleInstructionFactory(module)



import ast
from collections.abc import Collection
from types import ModuleType
from typing import final

import ast_comments

from cadwyn.codegen._common import CodegenContext


class LatestVersionAliasingPlugin:
    node_type = ast.Module

    @staticmethod
    def __call__(
        node: ast.Module,
        context: CodegenContext,
    ):
        if not context.current_version_is_latest:
            return node
        imports = _prepare_imports_from_version_dirs(
            context.template_module,
            ["latest"],
            context.index_of_latest_package_dir_in_module_python_path,
        )

        import_text = ast.unparse(imports[0].get_ast()) + " # noqa: F403"
        if context.extra["ignore_coverage_for_latest_aliases"]:
            import_text += " # pragma: no cover"
        return ast_comments.parse(import_text)


@final
class _ImportedModule:
    __slots__ = (
        "path",
        "name",
        "alias",
        "absolute_python_path_to_origin",
        "how_far_up_is_base_schema_dir_from_current_module",
        "is_package",
    )

    def __init__(
        self,
        version_dir: str,
        import_pythonpath_template: str,
        package_name: str,
        how_far_up_is_base_schema_dir_from_current_module: int,
        absolute_python_path_template: str,
        is_package: bool,
    ) -> None:
        self.path = import_pythonpath_template.format(version_dir)
        self.name = package_name.format(version_dir)
        if self.path == "":
            self.alias = self.name
        else:
            self.alias = f"{self.path.replace('.', '_')}_{self.name}"
        self.absolute_python_path_to_origin = absolute_python_path_template.format("latest")
        self.how_far_up_is_base_schema_dir_from_current_module = how_far_up_is_base_schema_dir_from_current_module
        self.is_package = is_package

    def get_ast(self) -> ast.ImportFrom:
        module = f"{self.path}.{self.name}"
        name = ast.alias(name="*")
        level = self.how_far_up_is_base_schema_dir_from_current_module
        if self.is_package and level == 2:
            level -= 1
        return ast.ImportFrom(
            level=level,
            module=module,
            names=[name],
        )


def _prepare_imports_from_version_dirs(
    original_module: ModuleType,
    version_dir_names: Collection[str],
    index_of_latest_package_dir_in_pythonpath: int,
) -> list[_ImportedModule]:
    # package.latest                     -> from .. import latest
    # package.latest.module              -> from ...latest import module
    # package.latest.subpackage          -> from ...latest import subpackage
    # package.latest.subpackage.module   -> from ....subpackage import module

    # package.latest                    -> from ..latest import *
    # package.latest.module             -> from ..latest.module import *
    # package.latest.subpackage         -> from ...latest.subpackage import *
    # package.latest.subpackage.module  -> from ...latest.subpackage.module import *

    original_module_parts = original_module.__name__.split(".")
    original_module_parts[index_of_latest_package_dir_in_pythonpath] = "{}"
    how_far_up_is_base_schema_dir_from_current_module = (
        len(original_module_parts) - index_of_latest_package_dir_in_pythonpath
    )
    is_package = original_module_parts[-1] == "__init__"
    if is_package:
        original_module_parts.pop(-1)

    package_name = original_module_parts[-1]
    package_path = original_module_parts[index_of_latest_package_dir_in_pythonpath:-1]
    import_pythonpath_template = ".".join(package_path)
    absolute_python_path_template = ".".join(original_module_parts)
    return [
        _ImportedModule(
            version_dir,
            import_pythonpath_template,
            package_name,
            how_far_up_is_base_schema_dir_from_current_module,
            absolute_python_path_template,
            is_package,
        )
        for version_dir in version_dir_names
    ]



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['cadwyn', 'cadwyn.codegen', 'cadwyn.codegen._plugins', 'cadwyn.structure']

package_data = \
{'': ['*']}

install_requires = \
['better-ast-comments>=1.2.1,<1.3.0',
 'fastapi>=0.96.1',
 'pydantic>=2.0.0,<3.0.0',
 'typing-extensions',
 'verselect>=0.1.1']

extras_require = \
{'cli': ['typer>=0.7.0']}

entry_points = \
{'console_scripts': ['cadwyn = cadwyn.__main__:app']}

setup_kwargs = {
    'name': 'cadwyn',
    'version': '3.3.1',
    'description': 'Production-ready community-driven modern Stripe-like API versioning in FastAPI',
    'long_description': '# Cadwyn\n\nProduction-ready community-driven modern [Stripe-like](https://stripe.com/blog/api-versioning) API versioning in FastAPI\n\n---\n\n<p align="center">\n<a href="https://github.com/zmievsa/cadwyn/actions?query=workflow%3ATests+event%3Apush+branch%3Amain" target="_blank">\n    <img src="https://github.com/zmievsa/cadwyn/actions/workflows/test.yaml/badge.svg?branch=main&event=push" alt="Test">\n</a>\n<a href="https://codecov.io/gh/ovsyanka83/cadwyn" target="_blank">\n    <img src="https://img.shields.io/codecov/c/github/ovsyanka83/cadwyn?color=%2334D058" alt="Coverage">\n</a>\n<a href="https://pypi.org/project/cadwyn/" target="_blank">\n    <img alt="PyPI" src="https://img.shields.io/pypi/v/cadwyn?color=%2334D058&label=pypi%20package" alt="Package version">\n</a>\n<a href="https://pypi.org/project/cadwyn/" target="_blank">\n    <img src="https://img.shields.io/pypi/pyversions/cadwyn?color=%2334D058" alt="Supported Python versions">\n</a>\n</p>\n\n## Who is this for?\n\nCadwyn allows you to support a single version of your code while auto-generating the schemas and routes for older versions. You keep API versioning encapsulated in small and independent "version change" modules while your business logic stays simple and knows nothing about versioning.\n\nIts [approach](https://docs.cadwyn.dev/theory/#ii-migration-based-response-building) will be useful if you want to:\n\n1. Support many API versions for a long time\n2. Effortlessly backport features and bugfixes to older API versions\n\nWhether you are a newbie in API versioning, a pro looking for a sophisticated tool, an experimenter looking to build a similar framework, or even someone who just wants to learn about all approaches to API versioning -- Cadwyn has the functionality, theory, and documentation to cover all the mentioned use cases.\n\n## Get started\n\nThe [documentation](https://docs.cadwyn.dev) has everything you need to succeed.\n\n## Sponsors\n\nThese are our gorgeous sponsors. They are using Cadwyn and are sponsoring it through various means. Contact [me](https://github.com/zmievsa) if you would like to become one too!\n\n[![Monite](https://docs.cadwyn.dev/img/sponsor_logos/monite.png)](https://docs.monite.com/)\n',
    'author': 'Stanislav Zmiev',
    'author_email': 'zmievsa@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/zmievsa/cadwyn',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)
